==> stacking_sats_pipeline-0.0.1.tar.gz: stacking_sats_pipeline-0.0.1/tests/test_backtest.py <==
#!/usr/bin/env python3
"""
Tests for stacking_sats_pipeline backtesting functionality
"""

from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from stacking_sats_pipeline import (
    BacktestResults,
    backtest,
    backtest_dynamic_dca,
    check_strategy_submission_ready,
    compute_cycle_spd,
    quick_backtest,
)
from stacking_sats_pipeline.backtest.runner import strategy


class TestBacktestingCore:
    """Test core backtesting functions."""

    def create_test_data(self, num_days=365):
        """Create test Bitcoin price data."""
        dates = pd.date_range("2020-01-01", periods=num_days, freq="D")
        np.random.seed(42)
        prices = 30000 + np.cumsum(np.random.normal(0, 500, num_days))
        prices = np.maximum(prices, 1000)
        return pd.DataFrame({"PriceUSD": prices}, index=dates)

    def create_simple_strategy(self):
        """Create a simple test strategy function."""

        def simple_strategy(df):
            """Equal weight strategy for testing."""
            base_weight = 1.0 / len(df)
            return pd.Series(base_weight, index=df.index)

        return simple_strategy

    def create_dca_strategy(self):
        """Create a simple DCA strategy."""

        def dca_strategy(df):
            """Buy more when price is below moving average."""
            df_copy = df.copy()
            ma_20 = df_copy["PriceUSD"].rolling(window=20, min_periods=1).mean()

            base_weight = 1.0 / len(df)
            weights = pd.Series(base_weight, index=df.index)

            # Buy 50% more when below moving average
            below_ma = df_copy["PriceUSD"] < ma_20
            weights[below_ma] *= 1.5

            return weights / weights.sum()

        return dca_strategy

    @pytest.mark.integration
    def test_quick_backtest_integration(self):
        """Integration test for quick_backtest (requires data loading)."""
        strategy_func = self.create_simple_strategy()

        try:
            result = quick_backtest(strategy_func)

            assert isinstance(result, (int, float))
            assert result > 0, (
                "Quick backtest should return positive performance metric"
            )

        except Exception as e:
            pytest.skip(f"Skipping integration test due to data issue: {e}")

    def test_strategy_decorator_basic(self):
        """Test the strategy decorator functionality."""

        @strategy(name="Test Strategy", auto_backtest=False)
        def test_strategy_func(df):
            return pd.Series(1.0 / len(df), index=df.index)

        # Check that decorator preserves function
        assert callable(test_strategy_func)

        # Check that decorator added metadata
        assert hasattr(test_strategy_func, "_strategy_name")
        assert test_strategy_func._strategy_name == "Test Strategy"

        # Test with sample data
        df = self.create_test_data(100)
        weights = test_strategy_func(df)

        assert isinstance(weights, pd.Series)
        assert len(weights) == len(df)
        assert np.isclose(weights.sum(), 1.0)

    @patch("stacking_sats_pipeline.backtest.runner.load_data")
    def test_backtest_mocked(self, mock_load_data):
        """Test backtest function with mocked data."""
        # Mock data loading
        mock_data = self.create_test_data(200)
        mock_load_data.return_value = mock_data

        strategy_func = self.create_simple_strategy()

        try:
            result = backtest(strategy_func)

            assert isinstance(result, BacktestResults)

        except Exception as e:
            # Backtest might fail due to other dependencies
            pytest.skip(f"Skipping mocked backtest due to: {e}")


class TestBacktestResults:
    """Test BacktestResults class functionality."""

    def test_backtest_results_creation(self):
        """Test BacktestResults can be instantiated."""
        # This test depends on the actual BacktestResults implementation
        try:
            # Try to create a BacktestResults instance
            # The exact parameters depend on the implementation
            results = BacktestResults()
            assert isinstance(results, BacktestResults)

        except TypeError:
            # If it requires parameters, skip this test
            pytest.skip("BacktestResults requires specific parameters")

    def test_backtest_results_methods(self):
        """Test that BacktestResults has expected methods."""
        import inspect

        # Check for common methods that should exist
        methods = inspect.getmembers(BacktestResults, predicate=inspect.isfunction)
        method_names = [name for name, _ in methods]

        # BacktestResults should have some analysis methods
        # The exact methods depend on implementation
        assert len(method_names) >= 0  # At least some methods should exist


class TestStrategyValidation:
    """Test strategy validation functions."""

    def create_test_data(self, num_days=365):
        """Create test data."""
        dates = pd.date_range("2020-01-01", periods=num_days, freq="D")
        np.random.seed(42)
        prices = 30000 + np.cumsum(np.random.normal(0, 500, num_days))
        prices = np.maximum(prices, 1000)
        return pd.DataFrame({"PriceUSD": prices}, index=dates)

    def test_check_strategy_submission_ready_valid(self):
        """Test check_strategy_submission_ready with valid strategy."""

        def valid_strategy(df):
            """A strategy that should pass validation."""
            base_weight = 1.0 / len(df)
            weights = pd.Series(base_weight, index=df.index)
            return weights / weights.sum()  # Ensure normalized

        df = self.create_test_data(200)

        try:
            # This function might print results or return validation status
            result = check_strategy_submission_ready(df, valid_strategy)
            # The function might not return anything, just validate

        except Exception as e:
            pytest.skip(f"Skipping validation test due to: {e}")

    def test_check_strategy_submission_ready_invalid(self):
        """Test check_strategy_submission_ready with invalid strategy."""

        def invalid_strategy(df):
            """A strategy that should fail validation (negative weights)."""
            weights = pd.Series(-1.0, index=df.index)  # Invalid negative weights
            return weights

        df = self.create_test_data(100)

        try:
            # This should either raise an exception or indicate failure
            result = check_strategy_submission_ready(df, invalid_strategy)

        except (ValueError, AssertionError):
            # Expected to fail with invalid strategy
            pass
        except Exception as e:
            pytest.skip(f"Skipping validation test due to: {e}")


class TestLegacyBacktesting:
    """Test legacy backtesting functions."""

    def create_test_data(self, num_days=365):
        """Create test data."""
        dates = pd.date_range("2020-01-01", periods=num_days, freq="D")
        np.random.seed(42)
        prices = 30000 + np.cumsum(np.random.normal(0, 500, num_days))
        prices = np.maximum(prices, 1000)
        return pd.DataFrame({"PriceUSD": prices}, index=dates)

    def test_compute_cycle_spd_basic(self):
        """Test compute_cycle_spd function."""

        def test_strategy(df):
            base_weight = 1.0 / len(df)
            return pd.Series(base_weight, index=df.index)

        df = self.create_test_data(400)  # Need enough data for cycles

        try:
            result = compute_cycle_spd(df, test_strategy)

            assert isinstance(result, pd.DataFrame)
            assert not result.empty

            # Check for expected columns (depends on implementation)
            expected_cols = ["dynamic_spd", "uniform_spd", "dynamic_pct", "uniform_pct"]
            available_cols = result.columns.tolist()

            # At least some SPD-related columns should exist
            assert len(available_cols) > 0

        except Exception as e:
            pytest.skip(f"Skipping compute_cycle_spd test due to: {e}")

    def test_backtest_dynamic_dca_basic(self):
        """Test backtest_dynamic_dca function."""

        def test_strategy(df):
            base_weight = 1.0 / len(df)
            return pd.Series(base_weight, index=df.index)

        df = self.create_test_data(400)

        try:
            # Capture output (function prints results)
            result = backtest_dynamic_dca(
                df, test_strategy, strategy_label="Test Strategy"
            )

            assert isinstance(result, pd.DataFrame)
            assert not result.empty

        except Exception as e:
            pytest.skip(f"Skipping backtest_dynamic_dca test due to: {e}")


class TestStrategyPatterns:
    """Test various strategy patterns and edge cases."""

    def test_constant_strategy(self):
        """Test strategy that always returns the same weights."""

        def constant_strategy(df):
            return pd.Series(1.0 / len(df), index=df.index)

        dates = pd.date_range("2020-01-01", periods=50, freq="D")
        prices = [30000] * 50
        df = pd.DataFrame({"PriceUSD": prices}, index=dates)

        weights = constant_strategy(df)

        assert isinstance(weights, pd.Series)
        assert len(weights) == len(df)
        assert np.allclose(weights, 1.0 / len(df))
        assert np.isclose(weights.sum(), 1.0)

    def test_varying_strategy(self):
        """Test strategy with time-varying weights."""

        def varying_strategy(df):
            """More weight at the beginning, less at the end."""
            weights = pd.Series(index=df.index, dtype=float)
            n = len(df)

            for i, idx in enumerate(df.index):
                # Linearly decreasing weights
                weights[idx] = (n - i) / (n * (n + 1) / 2)

            return weights

        dates = pd.date_range("2020-01-01", periods=10, freq="D")
        prices = list(range(30000, 30010))
        df = pd.DataFrame({"PriceUSD": prices}, index=dates)

        weights = varying_strategy(df)

        assert isinstance(weights, pd.Series)
        assert len(weights) == len(df)
        assert np.isclose(weights.sum(), 1.0, rtol=1e-10)
        assert (weights >= 0).all()

        # Should be decreasing
        weight_values = weights.values
        assert all(
            weight_values[i] >= weight_values[i + 1]
            for i in range(len(weight_values) - 1)
        )

==> stacking_sats_pipeline-0.2.0.tar.gz: stacking_sats_pipeline-0.2.0/tests/test_backtest.py <==
#!/usr/bin/env python3
"""
Tests for stacking_sats_pipeline backtesting functionality
"""

from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from stacking_sats_pipeline import (
    BacktestResults,
    backtest,
    backtest_dynamic_dca,
    check_strategy_submission_ready,
    compute_cycle_spd,
    quick_backtest,
)
from stacking_sats_pipeline.backtest.runner import strategy


class TestBacktestingCore:
    """Test core backtesting functions."""

    def create_test_data(self, num_days=365):
        """Create test Bitcoin price data."""
        dates = pd.date_range("2020-01-01", periods=num_days, freq="D")
        np.random.seed(42)
        prices = 30000 + np.cumsum(np.random.normal(0, 500, num_days))
        prices = np.maximum(prices, 1000)
        return pd.DataFrame({"PriceUSD": prices}, index=dates)

    def create_simple_strategy(self):
        """Create a simple test strategy function."""

        def simple_strategy(df):
            """Equal weight strategy for testing."""
            base_weight = 1.0 / len(df)
            return pd.Series(base_weight, index=df.index)

        return simple_strategy

    def create_dca_strategy(self):
        """Create a simple DCA strategy."""

        def dca_strategy(df):
            """Buy more when price is below moving average."""
            df_copy = df.copy()
            ma_20 = df_copy["PriceUSD"].rolling(window=20, min_periods=1).mean()

            base_weight = 1.0 / len(df)
            weights = pd.Series(base_weight, index=df.index)

            # Buy 50% more when below moving average
            below_ma = df_copy["PriceUSD"] < ma_20
            weights[below_ma] *= 1.5

            return weights / weights.sum()

        return dca_strategy

    @pytest.mark.integration
    @patch("stacking_sats_pipeline.backtest.runner.load_data")
    def test_quick_backtest_integration(self, mock_load_data):
        """Integration test for quick_backtest (requires data loading)."""
        # Mock data loading to avoid external dependencies
        mock_data = self.create_test_data(200)
        mock_load_data.return_value = mock_data

        strategy_func = self.create_simple_strategy()

        try:
            result = quick_backtest(strategy_func)

            assert isinstance(result, int | float)
            assert result > 0, "Quick backtest should return positive performance metric"

        except Exception as e:
            pytest.skip(f"Skipping integration test due to data issue: {e}")

    def test_strategy_decorator_basic(self):
        """Test the strategy decorator functionality."""

        @strategy(name="Test Strategy", auto_backtest=False)
        def test_strategy_func(df):
            return pd.Series(1.0 / len(df), index=df.index)

        # Check that decorator preserves function
        assert callable(test_strategy_func)

        # Check that decorator added metadata
        assert hasattr(test_strategy_func, "_strategy_name")
        assert test_strategy_func._strategy_name == "Test Strategy"

        # Test with sample data
        df = self.create_test_data(100)
        weights = test_strategy_func(df)

        assert isinstance(weights, pd.Series)
        assert len(weights) == len(df)
        assert np.isclose(weights.sum(), 1.0)

    @patch("stacking_sats_pipeline.backtest.runner.load_data")
    def test_backtest_mocked(self, mock_load_data):
        """Test backtest function with mocked data."""
        # Mock data loading
        mock_data = self.create_test_data(200)
        mock_load_data.return_value = mock_data

        strategy_func = self.create_simple_strategy()

        try:
            result = backtest(strategy_func)

            assert isinstance(result, BacktestResults)

        except Exception as e:
            # Backtest might fail due to other dependencies
            pytest.skip(f"Skipping mocked backtest due to: {e}")


class TestBacktestResults:
    """Test BacktestResults class functionality."""

    def create_test_data(self, num_days=365):
        """Create test data."""
        dates = pd.date_range("2020-01-01", periods=num_days, freq="D")
        np.random.seed(42)
        prices = 30000 + np.cumsum(np.random.normal(0, 500, num_days))
        prices = np.maximum(prices, 1000)
        return pd.DataFrame({"PriceUSD": prices}, index=dates)

    def test_backtest_results_creation(self):
        """Test BacktestResults can be instantiated."""
        try:
            # Create test data and strategy
            df = self.create_test_data(100)

            def test_strategy(data):
                return pd.Series(1.0 / len(data), index=data.index)

            # Create mock results
            results = {"spd_table": pd.DataFrame({"test": [1, 2, 3]})}

            # Create BacktestResults instance with proper parameters
            backtest_results = BacktestResults(test_strategy, df, results)
            assert isinstance(backtest_results, BacktestResults)

            # Test properties
            assert hasattr(backtest_results, "strategy_fn")
            assert hasattr(backtest_results, "df")
            assert hasattr(backtest_results, "results")
            assert hasattr(backtest_results, "weights")

        except Exception as e:
            pytest.skip(f"BacktestResults creation failed: {e}")

    def test_backtest_results_methods(self):
        """Test that BacktestResults has expected methods."""
        import inspect

        # Check for common methods that should exist
        methods = inspect.getmembers(BacktestResults, predicate=inspect.isfunction)
        method_names = [name for name, _ in methods]

        # BacktestResults should have some analysis methods
        # The exact methods depend on implementation
        assert len(method_names) >= 0  # At least some methods should exist


class TestStrategyValidation:
    """Test strategy validation functions."""

    def create_test_data(self, num_days=365):
        """Create test data."""
        dates = pd.date_range("2020-01-01", periods=num_days, freq="D")
        np.random.seed(42)
        prices = 30000 + np.cumsum(np.random.normal(0, 500, num_days))
        prices = np.maximum(prices, 1000)
        return pd.DataFrame({"PriceUSD": prices}, index=dates)

    def test_check_strategy_submission_ready_valid(self):
        """Test check_strategy_submission_ready with valid strategy."""

        def valid_strategy(df):
            """A strategy that should pass validation."""
            base_weight = 1.0 / len(df)
            weights = pd.Series(base_weight, index=df.index)
            return weights / weights.sum()  # Ensure normalized

        df = self.create_test_data(200)

        try:
            # This function might print results or return validation status
            check_strategy_submission_ready(df, valid_strategy)
            # The function might not return anything, just validate

        except Exception as e:
            pytest.skip(f"Skipping validation test due to: {e}")

    def test_check_strategy_submission_ready_invalid(self):
        """Test check_strategy_submission_ready with invalid strategy."""

        def invalid_strategy(df):
            """A strategy that should fail validation (negative weights)."""
            weights = pd.Series(-1.0, index=df.index)  # Invalid negative weights
            return weights

        df = self.create_test_data(100)

        try:
            # This should either raise an exception or indicate failure
            check_strategy_submission_ready(df, invalid_strategy)

        except (ValueError, AssertionError):
            # Expected to fail with invalid strategy
            pass
        except Exception as e:
            pytest.skip(f"Skipping validation test due to: {e}")


class TestLegacyBacktesting:
    """Test legacy backtesting functions."""

    def create_test_data(self, num_days=365):
        """Create test data."""
        dates = pd.date_range("2020-01-01", periods=num_days, freq="D")
        np.random.seed(42)
        prices = 30000 + np.cumsum(np.random.normal(0, 500, num_days))
        prices = np.maximum(prices, 1000)
        return pd.DataFrame({"PriceUSD": prices}, index=dates)

    def test_compute_cycle_spd_basic(self):
        """Test compute_cycle_spd function."""

        def test_strategy(df):
            base_weight = 1.0 / len(df)
            return pd.Series(base_weight, index=df.index)

        df = self.create_test_data(400)  # Need enough data for cycles

        try:
            result = compute_cycle_spd(df, test_strategy)

            assert isinstance(result, pd.DataFrame)
            assert not result.empty

            # Check for expected columns (depends on implementation)
            available_cols = result.columns.tolist()

            # At least some SPD-related columns should exist
            assert len(available_cols) > 0

        except Exception as e:
            pytest.skip(f"Skipping compute_cycle_spd test due to: {e}")

    def test_backtest_dynamic_dca_basic(self):
        """Test backtest_dynamic_dca function."""

        def test_strategy(df):
            base_weight = 1.0 / len(df)
            return pd.Series(base_weight, index=df.index)

        df = self.create_test_data(400)

        try:
            # Capture output (function prints results)
            result = backtest_dynamic_dca(df, test_strategy, strategy_label="Test Strategy")

            assert isinstance(result, pd.DataFrame)
            assert not result.empty

        except Exception as e:
            pytest.skip(f"Skipping backtest_dynamic_dca test due to: {e}")


class TestStrategyPatterns:
    """Test various strategy patterns and edge cases."""

    def test_constant_strategy(self):
        """Test strategy that always returns the same weights."""

        def constant_strategy(df):
            return pd.Series(1.0 / len(df), index=df.index)

        dates = pd.date_range("2020-01-01", periods=50, freq="D")
        prices = [30000] * 50
        df = pd.DataFrame({"PriceUSD": prices}, index=dates)

        weights = constant_strategy(df)

        assert isinstance(weights, pd.Series)
        assert len(weights) == len(df)
        assert np.allclose(weights, 1.0 / len(df))
        assert np.isclose(weights.sum(), 1.0)

    def test_varying_strategy(self):
        """Test strategy with time-varying weights."""

        def varying_strategy(df):
            """More weight at the beginning, less at the end."""
            weights = pd.Series(index=df.index, dtype=float)
            n = len(df)

            for i, idx in enumerate(df.index):
                # Linearly decreasing weights
                weights[idx] = (n - i) / (n * (n + 1) / 2)

            return weights

        dates = pd.date_range("2020-01-01", periods=10, freq="D")
        prices = list(range(30000, 30010))
        df = pd.DataFrame({"PriceUSD": prices}, index=dates)

        weights = varying_strategy(df)

        assert isinstance(weights, pd.Series)
        assert len(weights) == len(df)
        assert np.isclose(weights.sum(), 1.0, rtol=1e-10)
        assert (weights >= 0).all()

        # Should be decreasing
        weight_values = weights.values
        assert all(weight_values[i] >= weight_values[i + 1] for i in range(len(weight_values) - 1))
